using System;
using System.Collections.Generic;
using System.IdentityModel.Claims;
using System.IdentityModel.Policy;
using System.IdentityModel.Tokens;
using System.Security.Principal;
using System.ServiceModel;
using System.Text;
using SecurityTokenService;

namespace Service
{
    public class OperationAccessValidator : ServiceAuthorizationManager
    {
        protected override bool CheckAccessCore(OperationContext operationContext)
        {
            AuthorizationContext authContext =
                operationContext.ServiceSecurityContext.AuthorizationContext;

            // If there are no Claims in the AuthorizationContext, return false (Access Denied)
            // The issued token used to authenticate should contain claims 
            if(authContext.ClaimSets == null) return false;

            //
            // Action name is the actual service operation name
            //
            string actionName = operationContext.RequestContext.RequestMessage.Headers.Action;

            //
            // Caller identity name is specified using a Name Claim in the supplied token
            //
            string callerIdentityName = null;
            if(!TryGetClaimResourceFromClaimSet(authContext, ClaimTypes.Name, Rights.PossessProperty, out callerIdentityName))
                throw new SecurityTokenException("Icomming claim sets do not contain a name identity for the user.");

            //
            // Caller roles are also contained as claims in the supplied token
            //
            string[] callerRoles = LoadCallerRolesFromIncommingClaimSets(authContext);

            //
            // Create user Principal using identity name and roles
            //
            IPrincipal principal = CreatePrincipal(callerIdentityName, callerRoles);

            //
            // Set the principal property in order to WCF create the ThreadPrincipal later on
            //
            authContext.Properties["Principal"] = principal;

            //
            // Finally, validate user permissions to execute service operation based on user roles
            //
            return true; // when not using PrincipalPermission override this behavior to validate user credentials
        }

        #region Private Methods

        /// <summary>
        /// Creates a valid user principal
        /// </summary>
        /// <param name="callerIdentityName">User Identity name</param>
        /// <param name="callerRoles">User roles</param>
        /// <returns>User Principal object</returns>
        private IPrincipal CreatePrincipal(string callerIdentityName, string[] callerRoles)
        {
            GenericIdentity identity = new GenericIdentity(callerIdentityName, "SAMLToken");
            GenericPrincipal principal = new GenericPrincipal(identity, callerRoles);

            return principal;
        }

        /// <summary>
        /// Load user roles from current authorization context
        /// </summary>
        /// <param name="authContext"></param>
        /// <returns></returns>
        private string[] LoadCallerRolesFromIncommingClaimSets(AuthorizationContext authContext)
        {
            //
            // Iterate over supplied claim sets in search for role claims
            //
            List<string> callerRoleList = new List<string>();
            foreach(ClaimSet claimSet in authContext.ClaimSets)
                foreach(Claim claim in claimSet.FindClaims(Constants.RoleClaim, Rights.PossessProperty))
                    callerRoleList.Add((string)claim.Resource);

            return callerRoleList.ToArray();
        }

        /// <summary>
        /// Search a given claim type in the supplied claim sets and returns its resource
        /// </summary>
        /// <param name="authContext">Current Authorization Context</param>
        /// <param name="claimType">Claim Type to search for</param>
        /// <param name="rights">Type of right over the claim</param>
        /// <param name="resource">Returned claim resource</param>
        /// <returns>True if claim is found, False otherwise</returns>
        private static bool TryGetClaimResourceFromClaimSet(
            AuthorizationContext authContext,
            string claimType,
            string rights,
            out string resource)
        {
            resource = null;
            foreach(ClaimSet claimSet in authContext.ClaimSets)
                // Find all the Name claims and return the first valide claim
                foreach(Claim claim in claimSet)
                    if(claim.ClaimType == claimType &&
                        claim.Right == rights)
                    {
                        resource = (string)claim.Resource;
                        return true;
                    }

            return false;
        }

        #endregion
    }
}
